/*
 * Copyright (C) 2010 by egnite GmbH
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*!
 * \file httpmajs.c
 * \brief Minimal asynchronous JavaScript interface.
 *
 * \verbatim
 * $Id$
 * \endverbatim
 */

#include "config.h" /* Must be first. */

#if USE_HTTPSERVER

#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>

#include <dev/board.h>
#include <dev/urom.h>
#include <dev/vscodec.h>
#include <dev/watchdog.h>

#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include <sys/confos.h>
#include <sys/confnet.h>
#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/heap.h>
#include <sys/bankmem.h>
#include <sys/socket.h>
#include <sys/version.h>

#include <pro/ssi.h>
#include <pro/httpd.h>

#include <memdebug.h>

#include "logmsg.h"
#include "utils.h"
#include "xmlutils.h"
#include "favlist.h"
#include "proxy.h"
#include "webradio.h"
#include "shoutcast.h"
#include "httpserv.h"

#define CGI_PARAM_UPD_OS        0x01
#define CGI_PARAM_UPD_NET       0x02
#define CGI_PARAM_UPD_RADIO     0x04
#define CGI_PARAM_UPD_BOOT      0x80

typedef struct {
    char *s_search;
    char *s_genre;
    char *s_mt;
    char *s_br;
    int s_first;
    int s_last;
} DISCOVERY_PARMS;

static int SendFavorite(FILE *stream, char v, void *item);
static int SendGenre(FILE *stream, char v, void *item);
static int SendDiscovery(FILE *stream, char v, void *item);
static int SendGlobal(FILE *stream, char v, void *parms);

/*!
 * \brief Handle request for info page.
 *
 * Request is /cgi-bin/a_info.cgi.
 */
static int MajsSendTemplate(FILE * stream, char *name, void *parms, int list, void *item)
{
    int fh;
    int quoted = 0;
    int replace = 0;
    char *sp;
    char *dp;
    int icnt;
    char *ibuf;

    ibuf = malloc(128);
    if (ibuf == NULL) {
        return -1;
    }
    sprintf(ibuf, "UROM:templates/%s.inc", name);
    fh = _open(ibuf, _O_RDWR | _O_BINARY);
    if (fh == -1) {
        free(ibuf);
        return -1;
    }
    for (;;) {
        icnt = _read(fh, ibuf, sizeof(ibuf) - 1);
        if (icnt <= 0) {
            break;
        }
        ibuf[icnt] = '\0';
        sp = dp = ibuf;
        while (*sp) {
            if (!quoted && *sp >= '\x01' && *sp <= ' ') {
                sp++;
            }
            else if (*sp == '$') {
                if (replace) {
                    *dp++ = *sp++;
                    replace = 0;
                } else {
                    sp++;
                    replace = 1;
                    *dp = '\0';
                    fputs(ibuf, stream);
                    dp = ibuf;
                }
            }
            else if (replace) {
                switch (list) {
                case 1:
                    SendFavorite(stream, *sp, item);
                    break;
                case 2:
                    SendGenre(stream, *sp, item);
                    break;
                case 3:
                    SendDiscovery(stream, *sp, item);
                    break;
                default:
                    SendGlobal(stream, *sp, parms);
                    break;
                }
                sp++;
                replace = 0;
            }
            else {
                *dp++ = *sp;
                if (*sp++ == '"') {
                    quoted = !quoted;
                }
            }
        }
        *dp = '\0';
        fputs(ibuf, stream);
    }
    _close(fh);
    free(ibuf);

    return 0;
}

/*!
 * \brief Send HTTP header.
 *
 * Some browsers need a special header lines.
 */
static void SendHeader(FILE * stream, REQUEST * req)
{
    NutHttpSendHeaderTop(stream, req, 200, "Ok");
    fputs("Cache-Control: no-cache, must-revalidate\r\n", stream);
    fputs("Content-Type: text/html; charset=iso-8859-1\r\n", stream);
    fputs("Expires: Mon, 26 Jul 1997 05:00:00 GMT\r\n", stream);
    NutHttpSendHeaderBottom(stream, req, "text/html", -1);
}

/*!
 * \brief Send units with unit prefix.
 *
 * \param stream Stream to send to.
 * \param value  The number of units.
 * \param k      Size of the kilo prefix, typically 1024 for memory 
 *               sizes, 1000 otherwise.
 * \param unit   Unit name.
 */
static void SendUnit(FILE * stream, long value, long k, char *unit)
{
    char *s = " kMT";
    long d = 1L;

    while (*s) {
        if (value / d >= k) {
            d *= k;
            s++;
        } else {
            long ipart = value / d;
            fprintf(stream, "%lu.%04lu %c%s", ipart, (value - ipart * d) / 100UL, *s, unit);
            break;
        }
    }
}

/*!
 * \brief Check, if the given HTTP request is a post.
 *
 * \param req Pointer to the HTTP request structure.
 *
 * \return 0 if this request is a post. Otherwise -1 is returned.
 */
static int CheckForPost(REQUEST * req)
{
    if (req->req_method != METHOD_POST || req->req_type == NULL) {
        LogMsg(LOG_SHOUTCAST, "Bad method, POST expected\r\n");
        return -1;
    }
    if (strstr(req->req_type, "application/x-www-form-urlencoded") == NULL) {
        LogMsg(LOG_SHOUTCAST, "Bad content '%s'\r\n", req->req_type);
        return -1;
    }
    return 0;
}

/*!
 * \brief Collect given parameters from POST request.
 *
 * \param stream Stream to read from.
 * \param req    Pointer to the request structure.
 * \param names  This array must contain the parameter names, that 
 *               should be collected. The last element must be a
 *               NULL pointer.
 * \param values For each value found, a copy will be allocated
 *               from heap and the pointer will be stored in this
 *               array. The index into this array corresponds to 
 *               the names array. A trailing NULL pointer is not
 *               required.
 *
 * \return 0 on success, -1 otherwise.
 */
static int CollectPostedParms(FILE * stream, REQUEST * req, char *names[], char *values[])
{
    char *parm;
    int idx;

    /* Make sure this is a post. */
    if (CheckForPost(req)) {
        return -1;
    }

    /* Allocate the parameter buffer. */
    parm = malloc(2048);
    if (parm == NULL) {
        return -1;
    }

    /* Clear all values. */
    for (idx = 0; names[idx]; idx++) {
        values[idx] = NULL;
    }

    /* Collected the specified parameters. */
    for (;;) {
        /* Get the next parameter name. */
        if (GetEncodedString(stream, parm, 2048, '=') <= 0) {
            /* We reached the end. */
            break;
        }
        /* Check if this paramter is expected. */
        for (idx = 0; names[idx]; idx++) {
            if (strcmp(names[idx], parm) == 0) {
                break;
            }
        }
        /* We found a valid parameter, get its value. */
        if (GetEncodedString(stream, parm, 2048, '&') < 0) {
            break;
        }
        /* Store the new value. */
        if (names[idx]) {
            free(values[idx]);
            values[idx] = strdup(parm);
        }
    }

    /* Release the parameter buffer. */
    free(parm);

    return 0;
}

/*!
 * \brief Send the radio's info page.
 *
 * This function should be registered as a CGI routine.
 */
static int SendInfoPage(FILE * stream, REQUEST * req)
{
    SendHeader(stream, req);
    MajsSendTemplate(stream, "a_info", NULL, 0, NULL);
    fputs("\r\n", stream);
    fflush(stream);

    return 0;
}

/*!
 * \brief Send the info page values.
 *
 * This function should be registered as a CGI routine.
 */
static int SendInfoValues(FILE * stream, REQUEST * req)
{
    SHOUTCASTINFO *sci;
    static uint_fast8_t metaver;
    int wt;

    SendHeader(stream, req);

    /* Check for an update every half second, but do not wait longer than 5 seconds. */
    for (wt = 0; wt < 10; wt++) {
        if (webradio.wr_rip) {
            sci = (SHOUTCASTINFO *) webradio.wr_rip->ri_bcast;
            if (metaver != sci->sci_metaver) {
                metaver = sci->sci_metaver;
                break;
            }
        }
        NutSleep(500);
    }

    /* Send update. */
    MajsSendTemplate(stream, "av_info", NULL, 0, NULL);
    fputs("\r\n", stream);
    fflush(stream);

    return 0;
}

/*!
 * \brief Send the favorites page.
 *
 * This function should be registered as a CGI routine.
 */
static int SendFavoritesPage(FILE * stream, REQUEST * req)
{
    SendHeader(stream, req);
    MajsSendTemplate(stream, "a_favorites", NULL, 0, NULL);
    fputs("\r\n", stream);
    fflush(stream);

    return 0;
}

/*!
 * \brief Handle updates of the favorites list.
 *
 * This function should be registered as a CGI routine.
 */
static int FavoritesUpdate(FILE * stream, REQUEST * req)
{
    char *names[6] = { "action", "id", "title", "url", "newid", NULL};
    char *values[5];
    int rc;
    int i;

    /* Retrieve our parameters. */
    rc  = CollectPostedParms(stream, req, names, values);
    for (i = 0; names[i]; i++) {
        printf("%s=%s\n", names[i], values[i]);
    }

    /* Make sure we have at least an action and an ID. */
    if (values[0]) {
        if (values[1]) {
            int id = atoi(values[1]);

            if (strcmp(values[0], "save") == 0) {
                /* If we have a new name, save it. */
                if (values[2]) {
                    id = FavListSet(id, values[2], NULL);
                }
                /* If we have a new URLs, save'em. */
                if (values[3]) {
                    AddStationUris(id, values[3]);
                }
                /* Initiate non-volatile storage. */
                webradio.wr_favupd = 20;
            }
            else if (strcmp(values[0], "delete") == 0) {
                /* Delete this entry. */
                FavListSet(id, NULL, NULL);
                /* Initiate non-volatile storage. */
                webradio.wr_favupd = 20;
            }
            else if (strcmp(values[0], "play") == 0) {
                /* Copy this entry to the top. The main loop will
                automatically detect the change and switch to
                this station. */
                FavListCopy(id, TOP_FAVORITE);
            }
        }
        if (strcmp(values[0], "new") == 0) {
            if (values[2] && values[3]) {
                int id = FavListSet(MAX_FAVORITES, values[2], NULL);
                AddStationUris(id, values[3]);
                /* Send response with our id. */
                SendHeader(stream, req);
                fprintf(stream, "{\"id\":\"%d\",\"newid\":\"%s\"}\r\n", id, values[4]);
                fflush(stream);
                /* Initiate non-volatile storage. */
                webradio.wr_favupd = 20;
            }
        }
    }

    /* The collect routine allocated the value buffers from heap, 
       release them. */
    for (i = 0; names[i]; i++) {
        free(values[i]);
    }
    return rc;
}

/*!
 * \brief Send a variable value of the favorites list.
 */
static int SendFavorite(FILE *stream, char v, void *item)
{
    int i;
    int idx = *(int *)item;

    switch (v) {
    case 'e':
        /* Entry number. */
        fprintf(stream, "%d", idx);
        break;
    case 'n':
        /* Station name. */
        fputs(favlist[idx].rs_name, stream);
        break;
    case 'u':
        /* Station URLs. */
        for (i = 0; i < favlist[idx].rs_streams; i++) {
            fprintf(stream, "%s%%0A", favlist[idx].rs_uri[i]);
        }
        break;
    default:
        fprintf(stream, "%c?", v);
        break;
    }
    return 0;
}

/*!
 * \brief Send the favorites list.
 */
static int SendListOfFavorites(FILE *stream)
{
    int item;
    int first = 1;

    for (item = LAST_FAVORITE + 1; item < MAX_FAVORITES; item++) {
        if (favlist[item].rs_name == NULL) {
            continue;
        }
        if (first) {
            first = 0;
        } else {
            fputc(',', stream);
        }
        MajsSendTemplate(stream, "ai_favorites", NULL, 1, &item);
    }
    return 0;
}

/*!
 * \brief Send the discovery page.
 *
 * This function should be registered as a CGI routine.
 */
static int DiscoverPage(FILE * stream, REQUEST * req)
{
    SendHeader(stream, req);
    MajsSendTemplate(stream, "a_discover", NULL, 0, NULL);
    fputs("\r\n", stream);
    fflush(stream);

    return 0;
}

/*!
 * \brief Handle discovery page actions.
 *
 * This function should be registered as a CGI routine.
 */
static int DiscoverUpdate(FILE * stream, REQUEST * req)
{
    char *names[9] = { "action", "id", "search", "genre", "mt", "br", "from", "to", NULL};
    char *values[8];
    DISCOVERY_PARMS dscvr;
    int rc;
    int i;

    /* Retrieve our parameters. */
    rc  = CollectPostedParms(stream, req, names, values);

    if (rc == 0) {
        if (values[0]) {
            if (values[1]) {
                long id = atol(values[1]);
                if (strcmp(values[0], "play") == 0) {
                    /* Add this entry to the top. The main loop will
                    automatically detect the change and switch to
                    this station. */
                    ShoutCastAddStationById(id, TOP_FAVORITE);
                }
                else if (strcmp(values[0], "add") == 0) {
                    /* Add this entry to the favorites list. */
                    ShoutCastAddStationById(id, MAX_FAVORITES);
                    /* Initiate non-volatile storage. */
                    webradio.wr_favupd = 20;
                }
            }
        } else {
            dscvr.s_search = values[2];
            dscvr.s_genre = values[3];
            dscvr.s_mt = values[4];
            dscvr.s_br = values[5];
            dscvr.s_first = values[6] ? atoi(values[6]) : 0;
            dscvr.s_last = values[7] ? atoi(values[7]) : 999;

            /* Send the response template. */
            SendHeader(stream, req);
            MajsSendTemplate(stream, "as_discover", &dscvr, 0, NULL);
            fputs("\r\n", stream);
            fflush(stream);
        }
    }

    /* The collect routine allocated the value buffers from heap, 
       release them. */
    for (i = 0; names[i]; i++) {
        free(values[i]);
    }
    return rc;
}

/*!
 * \brief Send a variable value of the discovery page.
 */
static int SendDiscovery(FILE *stream, char v, void *item)
{
    char **station = (char **)item;

    switch (v) {
    case 'n':
        /* Station name. */
        fprintf(stream, "%s", station[0]);
        break;
    case 't':
        /* Mediadata type. */
        fprintf(stream, "%s", station[1]);
        break;
    case 'i':
        /* Station ID. */
        fprintf(stream, "%s", station[2]);
        break;
    case 'r':
        /* Bitrate. */
        fprintf(stream, "%s", station[3]);
        break;
    case 'g':
        /* Genre. */
        fprintf(stream, "%s", station[4]);
        break;
    case 'c':
        /* Current title. */
        fprintf(stream, "%s", station[5]);
        break;
    case 'l':
        /* Number of listeners. */
        fprintf(stream, "%s", station[6]);
        break;
    default:
        fprintf(stream, "%c?", v);
        break;
    }
    return 0;
}

/*!
 * \brief Do a SHOUTcast search and send a list of results.
 */
static int SendListOfDiscoveries(FILE *stream, void *parms)
{
    XMLSTREAM *xs;
    char *query;
    char *qp;
    char *sep = "?";
    int idx;
    DISCOVERY_PARMS *dscvr = (DISCOVERY_PARMS *) parms;

    query = malloc(256);
    if (query == NULL) {
        return -1;
    }
    strcpy(query, SHOUTCAST_QLIST);
    qp = query + strlen(query);
    if (dscvr->s_genre && strcmp(dscvr->s_genre, "any")) {
        strcat(query, sep);
        sep = "&";
        strcat(query, "genre=");
        strcat(query, dscvr->s_genre);
    }
    if (dscvr->s_search) {
        strcat(query, sep);
        sep = "&";
        strcat(query, "search=");
        strcat(query, dscvr->s_search);
    }
    if (dscvr->s_mt && strcmp(dscvr->s_mt, "any")) {
        strcat(query, sep);
        sep = "&";
        strcat(query, "mt=");
        strcat(query, dscvr->s_mt);
    }
    if (dscvr->s_br && atoi(dscvr->s_br)) {
        strcat(query, sep);
        sep = "&";
        strcat(query, "br=");
        strcat(query, dscvr->s_br);
    }

    xs = XmlStreamOpen(query);
    free(query);

    if (xs) {

        int first = 1;
        idx = 0;
        for (;;) {
            int invalid = 0;
            int i;
            char *attr[7] = { "name", "mt", "id", "br", "genre", "ct", "lc" };
            char **station = XmlStreamGet(xs, "station", attr, 7);

            if (station == NULL) {
                break;
            }
            for (i = 0; i < 7; i++) {
                if (strchr(station[i], '\n')) {
                    invalid = 1;
                }
            }

            if (++idx >= dscvr->s_first && !invalid) {
                if (first) {
                    first = 0;
                } else {
                    fputc(',', stream);
                }
                MajsSendTemplate(stream, "ai_discover", NULL, 3, station);

            }
            for (i = 0; i < 7; i++) {
                free(station[i]);
            }
            free(station);
            if (idx > dscvr->s_last) {
                break;
            }
        }
        XmlStreamClose(xs);
    }
    return 0;
}

/*!
 * \brief Send a variable value of the genre list.
 */
static int SendGenre(FILE *stream, char v, void *item)
{
    char *genre = (char *)item;

    switch (v) {
    case 'g':
        /* Entry number. */
        fprintf(stream, "%s", genre);
        break;
    default:
        fprintf(stream, "%c?", v);
        break;
    }
    return 0;
}

/*!
 * \brief Send the genre list.
 */
static int SendListOfGenres(FILE *stream)
{
    int first = 1;
    XMLSTREAM *xs;

    xs = XmlStreamOpen(SHOUTCAST_QLIST);
    if (xs) {
        for (;;) {
            char *attr[1] = { "name" };
            char **genre = XmlStreamGet(xs, "genre", attr, 1);

            if (genre == NULL) {
                break;
            }
            if (strchr(genre[0], '\n') == NULL) {
                if (first) {
                    first = 0;
                } else {
                    fputc(',', stream);
                }
                MajsSendTemplate(stream, "ai_genres", NULL, 2, genre[0]);
            }
            free(genre[0]);
            free(genre);
        }
        XmlStreamClose(xs);
    }
    return 0;
}

/*!
 * \brief Send the settings page.
 *
 * This function should be registered as a CGI routine.
 */
static int SettingsPage(FILE * stream, REQUEST * req)
{
    SendHeader(stream, req);
    MajsSendTemplate(stream, "a_settings", NULL, 0, NULL);
    fputs("\r\n", stream);
    fflush(stream);

    return 0;
}

/*!
 * \brief Handle settings page actions.
 *
 * This function should be registered as a CGI routine.
 */
static int SettingsUpdate(FILE * stream, REQUEST * req)
{
    int rc;
    char *names[10] = { "submit", "hostname", "proxy", "dhcp", "ip", "mask", "gateway", "dns0", "dns1", NULL};
    char *values[9];
    uint_fast8_t update = 0;
    int i;

    /* Retrieve our parameters. */
    rc  = CollectPostedParms(stream, req, names, values);
    if (rc == 0 && values[0]) {
        if (strcmp(values[0], "save") == 0) {
            if (values[1] && *values[1] && strcmp(values[1], confos.hostname)) {
                strncpy(confos.hostname, values[1], sizeof(confos.hostname) - 1);
                update |= CGI_PARAM_UPD_OS;
            }
            if (values[2]) {
                memset(&proxy, 0, sizeof(proxy));
                if (*values[2]) {
                    char *cp = strrchr(proxy.proxy_host, ':');
                    strncpy(proxy.proxy_host, values[2], sizeof(proxy.proxy_host) - 1);
                    if (cp) {
                        *cp++ = 0;
                        proxy.proxy_port = atoi(cp);
                    }
                }
                update |= CGI_PARAM_UPD_RADIO;
            }
            if (values[3] && atoi(values[3])) {
                confnet.cdn_cip_addr = 0;
                update |= CGI_PARAM_UPD_NET;
            } else {
                uint32_t addr;

                if (values[4]) {
                    addr = inet_addr(values[4]);
                    if (addr != (uint32_t) -1 && addr != confnet.cdn_cip_addr) {
                        confnet.cdn_cip_addr = addr;
                        update |= CGI_PARAM_UPD_NET | CGI_PARAM_UPD_BOOT;
                    }
                }
                if (values[5]) {
                    addr = inet_addr(values[5]);
                    if (addr != (uint32_t) -1 && addr != confnet.cdn_ip_mask) {
                        confnet.cdn_ip_mask = addr;
                        update |= CGI_PARAM_UPD_NET | CGI_PARAM_UPD_BOOT;
                    }
                }
                if (values[6]) {
                    addr = inet_addr(values[6]);
                    if (addr != (uint32_t) -1 && addr != confnet.cdn_gateway) {
                        confnet.cdn_gateway = addr;
                        update |= CGI_PARAM_UPD_NET | CGI_PARAM_UPD_BOOT;
                    }
                }
                if (values[7]) {
                    addr = inet_addr(values[7]);
                    if (addr != (uint32_t) -1 && addr != webradio.wr_pridns) {
                        webradio.wr_pridns = addr;
                        update |= CGI_PARAM_UPD_RADIO;
                    }
                }
                if (values[8]) {
                    addr = inet_addr(values[8]);
                    if (addr != (uint32_t) -1 && addr != webradio.wr_secdns) {
                        webradio.wr_secdns = addr;
                        update |= CGI_PARAM_UPD_RADIO;
                    }
                }
            }
        }
        else if (strcmp(values[0], "erase") == 0) {
        }
        if (update & CGI_PARAM_UPD_OS) {
            LogMsg(LOG_HTTPD, "Saving OS\r\n");
            NutSaveConfig();
        }
        if (update & CGI_PARAM_UPD_NET) {
            LogMsg(LOG_HTTPD, "Saving Net\r\n");
            NutNetSaveConfig();
        }
        if (cfgFile) {
            if (update & CGI_PARAM_UPD_RADIO) {
                ConfigSave(cfgFile);
            }
        }
        if (update & CGI_PARAM_UPD_BOOT) {
            LogMsg(LOG_HTTPD, "Reboot\r\n");
            webradio.wr_reboot = 3;
        }
    }

    /* The collect routine allocated the value buffers from heap, 
       release them. */
    for (i = 0; names[i]; i++) {
        free(values[i]);
    }
    return rc;

}

/*!
 * \brief Send a variable value.
 */
static int SendGlobal(FILE *stream, char v, void *parms)
{
    switch (v) {
    case '1':
        /* List of favorites. */
        SendListOfFavorites(stream);
        break;
    case '2':
        /* List of genres. */
        SendListOfGenres(stream);
        break;
    case '3':
        /* List of discovered stations. */
        SendListOfDiscoveries(stream, parms);
        break;

    case 'A':
        /* Configured IP address. */
        fputs(inet_ntoa(confnet.cdn_cip_addr ? confnet.cdn_cip_addr : confnet.cdn_ip_addr), stream);
        break;
    case 'b':
        /* Buffer used. */
        SendUnit(stream, NutSegBufUsed(), 1024, "Bytes");
        break;
    case 'C':
        /* CPU clock. */
        SendUnit(stream, NutGetCpuClock(), 1000, "Hz");
        break;
    case 'D':
        /* DHCP configured. */
        fputc(confnet.cdn_cip_addr ? '0' : '1', stream);
        break;
    case 'G':
        /* Configured gateway IP. */
        fputs(inet_ntoa(confnet.cdn_gateway), stream);
        break;
    case 'g':
        /* Current station genre. */
        if (webradio.wr_sip && webradio.wr_sip->si_genre) {
            fputs(webradio.wr_sip->si_genre, stream);
        }
        break;
    case 'H':
        /* Configured host name. */
        fputs(confos.hostname, stream);
        break;
    case 'i':
        /* Stream metadata interval. */
        if (webradio.wr_rip) {
            SHOUTCASTINFO *sci = (SHOUTCASTINFO *) webradio.wr_rip->ri_bcast;
            SendUnit(stream, sci->sci_metaint, 1024, "Bytes");
        }
        break;
    case 'm':
        /* Memory used. */
        {
            long mem = NUTMEM_SIZE;
#if NUTXMEM_SIZE
            mem += NUTXMEM_SIZE;
#endif
            SendUnit(stream, mem - NutHeapAvailable(), 1024, "Bytes");
        }
        break;
    case 'M':
        /* Configured IP mask. */
        fputs(inet_ntoa(confnet.cdn_ip_mask), stream);
        break;
    case 'n':
        /* Current station name. */
        if (webradio.wr_sip) {
            fputs(webradio.wr_sip->si_name, stream);
        }
        break;
    case 'P':
        /* Configured primary DNS. */
        {
            uint32_t addr;
            NutDnsGetConfig2(NULL, NULL, &addr, NULL);
            fputs(inet_ntoa(addr), stream);
        }
        break;
    case 'r':
        /* Current bitrate. */
        if (webradio.wr_sip) {
            fprintf(stream, "%u kBit", webradio.wr_sip->si_bitrate);
        }
        break;
    case 'S':
        /* Configured secondary DNS. */
        {
            uint32_t addr;
            NutDnsGetConfig2(NULL, NULL, NULL, &addr);
            fputs(inet_ntoa(addr), stream);
        }
        break;
    case 't':
        /* Stream metadata title. */
        if (webradio.wr_rip) {
            SHOUTCASTINFO *sci = (SHOUTCASTINFO *) webradio.wr_rip->ri_bcast;
            if (sci && sci->sci_metatitle) {
                fputs(sci->sci_metatitle, stream);
            }
        }
        break;
    case 'T':
        /* Master clock. */
        SendUnit(stream, NutClockGet(NUT_HWCLK_PERIPHERAL), 1000, "Hz");
        break;
    case 'u':
        /* Stream metadata URL. */
        if (webradio.wr_rip) {
            SHOUTCASTINFO *sci = (SHOUTCASTINFO *) webradio.wr_rip->ri_bcast;
            if (sci->sci_metaurl) {
                fputs(sci->sci_metaurl, stream);
            }
        }
        break;
    case 'X':
        /* Configured proxy IP. */
        if (proxy.proxy_port) {
            fputs(proxy.proxy_host, stream);
            if (proxy.proxy_port != 80) {
                fprintf(stream, ":%u", proxy.proxy_port);
            }
        }
        break;
    default:
        fprintf(stream, "%c?", v);
        break;
    }
    return 0;
}

/*!
 * \brief Register CGIs used by JavaScript.
 *
 * - cgi-bin/a_info.cgi Sends info page template.
 * - cgi-bin/av_info.cgi Sends dynamic values for info page.
 * - cgi-bin/a_favorites.cgi Sends favorites page template.
 * - cgi-bin/as_favorites.cgi Handles user updates on favorites page.
 * - cgi-bin/a_discover.cgi Sends discovery page template.
 * - cgi-bin/as_discover.cgi Handles search function on discovery page.
 * - cgi-bin/a_settings.cgi Sends settings page template.
 * - cgi-bin/as_settings.cgi Handles user updates on settings page.
 */
int RegisterMajs(void)
{
    NutRegisterCgi("a_info.cgi", SendInfoPage);
    NutRegisterCgi("av_info.cgi", SendInfoValues);
    NutRegisterCgi("a_favorites.cgi", SendFavoritesPage);
    NutRegisterCgi("as_favorites.cgi", FavoritesUpdate);
    NutRegisterCgi("a_discover.cgi", DiscoverPage);
    NutRegisterCgi("as_discover.cgi", DiscoverUpdate);
    NutRegisterCgi("a_settings.cgi", SettingsPage);
    NutRegisterCgi("as_settings.cgi", SettingsUpdate);

    return 0;
}

#endif /* USE_HTTPSERVER */
